local _G = getfenv(0)
local gratuity = AceLibrary("Gratuity-2.0")

ElkBuffBars = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0")
ElkBuffBars:RegisterDB("ElkBuffBarsDB")

local STICKTO_AREA = 25

local updates_disabled = false
local scan_happend = {}
scan_happend.player = false
scan_happend.pet = false
scan_happend.target = false
local old_hasMainHandEnchant, old_mainHandExpiration, old_mainHandCharges, old_hasOffHandEnchant, old_offHandExpiration, old_offHandCharges

function ElkBuffBars:OnInitialize()
	local build = select(2, _G.GetBuildInfo())
	if not self.db.account.build or self.db.account.build ~= build then
		self.db.account.build = build
		self.db.account.maxtimes = {
			BUFF = {},
			DEBUFF = {},
			TENCH = {},
		}
	end
end

function ElkBuffBars:OnEnable()
	-- create bargroups based on stored settings
	self:CreateBarGroups()
	
	-- register events
	self:RegisterBucketEvent("PLAYER_AURAS_CHANGED", .25)
	self:RegisterBucketEvent("PLAYER_TARGET_CHANGED", .25)
	self:RegisterBucketEvent("UNIT_AURA", .25)
	self:ScheduleRepeatingEvent(self.DoUpdates, .1, self)
	
	--
	self:ScanData_BUFF()
	self:ScanData_DEBUFF()
	self:ScanData_TENCH()
	self:ScanData_TRACKING()
end

function ElkBuffBars:OnDisable()
	-- recycle all bargroups
	self:RemoveBarGroups()
end


--
--
--
local barcache = {}

function ElkBuffBars:GetBar()
	if #barcache > 0 then
		return table.remove(barcache, #barcache)
	else
		return EBB_Bar:new()
	end
end

function ElkBuffBars:RecycleBar(bar)
	bar:Reset()
	table.insert(barcache, bar)
end

--
--
--
local bargroupcache = {}

function ElkBuffBars:GetBarGroup()
	if #bargroupcache > 0 then
		return table.remove(bargroupcache, #bargroupcache)
	else
		return EBB_BarGroup:new()
	end
end

function ElkBuffBars:RecycleBarGroup(bargroup)
	bargroup:Reset()
	table.insert(bargroupcache, bargroup)
end

--
--
--
local datatablecache = {}

local function GetDataTable()
	if #datatablecache > 0 then
		return table.remove(datatablecache, #datatablecache)
	else
		return {}
	end
end

local function RecycleDataTable(dt)
	table.insert(datatablecache, dt)
end

--
--
--
function ElkBuffBars:CreateBarGroups()
	if not self.bargroups then
		self.bargroups = {}
	else
		for k, v in pairs(self.bargroups) do
			self:RecycleBarGroup(v)
			self.bargroups[k] = nil
		end
	end

-- -----
	self.db.profile.bargroups = {}
	local layout = {
		["bars"] = {
			["icon"]				= "LEFT",											-- "LEFT", "RIGHT", false
			["iconcount"]			= true,												-- true, false
			["iconcountanchor"]		= "CENTER",											-- <anchor>
			["iconcountfont"]		= "Arial Narrow",									-- <SML:font>
			["iconcountfontsize"]	= 14,												-- <font size>
			["iconcountcolor"]		= {1, 1, 1},										-- <color set>
			["bar"]					= "Otravi",											-- <SML:statusbar>, false
			["barright"]			= false,											-- true, false
			["textTL"]				= "NAMERANKCOUNT",									-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTLfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textTLfontsize"]		= 14,												-- <font size>
			["textTLcolor"]			= {1, 1, 1},										-- <color set>
			["textTLalign"]			= "LEFT",											-- left, center, right
			["textTR"]				= "TIMELEFT",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTRfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textTRfontsize"]		= 14,												-- <font size>
			["textTRcolor"]			= {1, 1, 1},										-- <color set>
			["barcolor"]			= {0.3, 0.5, 1},									-- <color set>
			["barbgcolor"]			= {0, 0.5, 1},										-- <color set>
			["debufftypecolor"]		= true,												-- true, false
			["timeformat"]			= "CONDENSED",										-- "DEFAULT", "EXTENDED", "FULL", "SHORT", "CONDENSED"
			["width"]				= 350,
			["height"]				= 25,
		},
		["filter"] = {
			["type"] = {
				["BUFF"] = true,
				["TENCH"] = true,
				["TRACKING"] = true,
			}
		},
		sorting = "timeleft",						-- timeleft, timemax
		target = "player",							-- player, pet, target
		growup = false,								-- true, false
		barspacing = 0,								-- 0+
		anchortext = "BUFFS & Co.",					-- <string>
		anchorshown = false,						-- true, false
	}
	table.insert(self.db.profile.bargroups, layout)
	layout = {
		["bars"] = {
			["icon"]				= "LEFT",											-- "LEFT", "RIGHT", false
			["iconcount"]			= true,												-- true, false
			["iconcountanchor"]		= "BOTTOMRIGHT",									-- <anchor>
			["iconcountfont"]		= "Arial Narrow",									-- <SML:font>
			["iconcountfontsize"]	= 14,												-- <font size>
			["iconcountcolor"]		= {1, 1, 1},										-- <color set>
			["bar"]					= "Charcoal",										-- <SML:statusbar>, false
			["barright"]			= true,												-- true, false
			["textTL"]				= "NAMERANK",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTLfont"]			= "Morpheus",										-- <SML:font>
			["textTLfontsize"]		= 20,												-- <font size>
			["textTLcolor"]			= {1, 1, 1},										-- <color set>
			["textTLalign"]			= "LEFT",											-- left, center, right
			["textTR"]				= "COUNT",											-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTRfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textTRfontsize"]		= 14,												-- <font size>
			["textTRcolor"]			= {1, 1, 1},										-- <color set>
			["textBL"]				= "DEBUFFTYPE",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textBLfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textBLfontsize"]		= 14,												-- <font size>
			["textBLcolor"]			= {1, 1, 1},										-- <color set>
			["textBLalign"]			= "LEFT",											-- left, center, right
			["textBR"]				= "TIMELEFT",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textBRfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textBRfontsize"]		= 14,												-- <font size>
			["textBRcolor"]			= {1, 1, 1},										-- <color set>
			["barcolor"]			= {1, 0, 0},										-- <color set>
			["barbgcolor"]			= {1, 0, 0},										-- <color set>
			["debufftypecolor"]		= true,												-- true, false
			["timeformat"]			= "CONDENSED",										-- "DEFAULT", "EXTENDED", "FULL", "SHORT", "CONDENSED"
			["width"]				= 350,
			["height"]				= 50,
		},
		["filter"] = {
			["type"] = {
				["DEBUFF"] = true,
			}
		},
		sorting = "timeleft",						-- timeleft, timemax
		target = "player",							-- player, pet, target
		growup = false,								-- true, false
		barspacing = 0,								-- 0+
		anchortext = "DEBUFFS",						-- <string>
		anchorshown = false,						-- true, false
		stickto = 1,								-- bargroup id
		stickside = "",								-- "LEFT", "RIGHT", ""
	}
	table.insert(self.db.profile.bargroups, layout)
	layout = {
		["bars"] = {
			["icon"]				= "LEFT",											-- "LEFT", "RIGHT", false
			["iconcount"]			= true,												-- true, false
			["iconcountanchor"]		= "BOTTOMRIGHT",									-- <anchor>
			["iconcountfont"]		= "Arial Narrow",									-- <SML:font>
			["iconcountfontsize"]	= 14,												-- <font size>
			["iconcountcolor"]		= {1, 1, 1},										-- <color set>
			["bar"]					= "Steel",											-- <SML:statusbar>, false
			["barright"]			= false,											-- true, false
			["textTL"]				= "NAMERANK",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTLfont"]			= "Morpheus",										-- <SML:font>
			["textTLfontsize"]		= 20,												-- <font size>
			["textTLcolor"]			= {1, 1, 1},										-- <color set>
			["textTLalign"]			= "LEFT",											-- left, center, right
			["textTR"]				= "COUNT",											-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTRfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textTRfontsize"]		= 14,												-- <font size>
			["textTRcolor"]			= {1, 1, 1},										-- <color set>
			["textBL"]				= "DEBUFFTYPE",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textBLfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textBLfontsize"]		= 14,												-- <font size>
			["textBLcolor"]			= {1, 1, 1},										-- <color set>
			["textBLalign"]			= "LEFT",											-- left, center, right
			["textBR"]				= "TIMELEFT",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textBRfont"]			= "Friz Quadrata TT",								-- <SML:font>
			["textBRfontsize"]		= 14,												-- <font size>
			["textBRcolor"]			= {1, 1, 1},										-- <color set>
			["barcolor"]			= {1, 0, 0},										-- <color set>
			["barbgcolor"]			= {1, 0, 0},										-- <color set>
			["debufftypecolor"]		= true,												-- true, false
			["timeformat"]			= "CONDENSED",										-- "DEFAULT", "EXTENDED", "FULL", "SHORT", "CONDENSED"
			["width"]				= 350,
			["height"]				= 50,
		},
		["filter"] = {
			["type"] = {
				["TARGETBUFF"] = true,
				["TARGETDEBUFF"] = true,
			}
		},
		sorting = "timeleft",						-- timeleft, timemax
		target = "target",							-- player, pet, target
		growup = false,								-- true, false
		barspacing = 0,								-- 0+
		anchortext = "Target...",					-- <string>
		anchorshown = false,						-- true, false
	}
	table.insert(self.db.profile.bargroups, layout)
-- -----
	local bglayouts = self.db.profile.bargroups
	for k, v in ipairs(bglayouts) do
		self:AddBarGroup(v)
	end
	for k, v in ipairs(self.bargroups) do
		local layout = bglayouts[k]
		local container = v:GetContainer()
		container:Show()
		if layout.stickto then
			container:SetPoint((layout.growup and "BOTTOM" or "TOP")..(layout.stickside or ""), self.bargroups[layout.stickto]:GetContainer(), (layout.growup and "TOP" or "BOTTOM")..(layout.stickside or ""))
		elseif layout.x and layout.y then
			container:SetPoint(layout.growup and "BOTTOMLEFT" or "TOPLEFT", UIParent, "BOTTOMLEFT", layout.x, layout.y)
		else
			container:SetPoint("CENTER", UIParent, "CENTER", layout.x, layout.y)
			v:ToggleAnchor(true)
		end
	end
end

function ElkBuffBars:RemoveBarGroups()
	for k, v in self.bargroups do
		self:RecycleBarGroup(v)
		self.bargroups[k] = nil
	end
end

function ElkBuffBars:AddBarGroup(layout)
	if not layout then
		layout = {
		}
		table.insert(self.db.profile.bargroups, layout)
	end
	local bargroup = self:GetBarGroup()
	table.insert(self.bargroups, bargroup)
	layout.id = #self.bargroups
	bargroup:SetLayout(layout)
end

function ElkBuffBars:RemoveBarGroup(id)
	table.remove(self.bargroups, id)
	for k, v in self.bargroups do
		local layout = v.layout
		layout.id = k
		if layout.stickto then
			if layout.stickto == id then
				-- the group we sticked to was removed
				local container = v:GetContainer()
				layout.stickto = nil
				container:ClearAllPoints()
				container:SetPoint("CENTER", UIParent, "CENTER", layout.x, layout.y)
				v:ToggleAnchor(true)
			elseif layout.stickto > id then
				layout.stickto = layout.stickto - 1
			end
		end
	end
end

--
--
--
ElkBuffBars.buffdata = {}
ElkBuffBars.debuffdata = {}
ElkBuffBars.tenchdata = {}
ElkBuffBars.trackingdata = {}
ElkBuffBars.targetbuffdata = {}
ElkBuffBars.targetdebuffdata = {}

function ElkBuffBars:PLAYER_AURAS_CHANGED()
	self:ScanData_BUFF()
	self:ScanData_DEBUFF()
	self:ScanData_TRACKING()
end

function ElkBuffBars:PLAYER_TARGET_CHANGED()
	self:ScanData_Unit_BUFF("target")
	self:ScanData_Unit_DEBUFF("target")
end

function ElkBuffBars:UNIT_AURA(args)
	local scanned_pet = false
	local scanned_target = false
	for arg in pairs(args) do
		if not scanned_pet and arg == "pet" then
			self:ScanData_Unit_BUFF("pet")
			self:ScanData_Unit_DEBUFF("pet")
		end
		if not scanned_target and arg == "target" then
			self:ScanData_Unit_BUFF("target")
			self:ScanData_Unit_DEBUFF("target")
		end
	end
end

function ElkBuffBars:ScanData_BUFF()
	updates_disabled = true
	local maxtimes = self.db.account.maxtimes.BUFF
	for k, v in pairs(self.buffdata) do
		RecycleDataTable(v)
		self.buffdata[k] = nil
	end
	local i = 1
	while true do
		local buffIndex, untilCancelled = GetPlayerBuff(i, "HELPFUL")
		if buffIndex == 0 then break end
		local timeleft = GetPlayerBuffTimeLeft(buffIndex)
		local timemax = timeleft
		local name, rank = GetPlayerBuffName(buffIndex)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end

		local dt = GetDataTable()
		dt.id				= buffIndex
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "BUFF"
		dt.realtype			= "BUFF"
		dt.debufftype		= nil
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= (untilCancelled == 1) and true or nil
		dt.charges			= GetPlayerBuffApplications(buffIndex)
		dt.icon				= GetPlayerBuffTexture(buffIndex)
		
		table.insert(self.buffdata, dt)
		i = i + 1
	end
	scan_happend.player = true
	updates_disabled = false
end

function ElkBuffBars:ScanData_DEBUFF()
	updates_disabled = true
	local maxtimes = self.db.account.maxtimes.DEBUFF
	for k, v in pairs(self.debuffdata) do
		RecycleDataTable(v)
		self.debuffdata[k] = nil
	end
	local i = 1
	while true do
		local buffIndex, untilCancelled = GetPlayerBuff(i, "HARMFUL")
		if buffIndex == 0 then break end
		local timeleft = GetPlayerBuffTimeLeft(buffIndex)
		local timemax = timeleft
		local name, rank = GetPlayerBuffName(buffIndex)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end

		local dt = GetDataTable()
		dt.id				= buffIndex
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "DEBUFF"
		dt.realtype			= "DEBUFF"
		dt.debufftype		= GetPlayerBuffDispelType(buffIndex)
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= (untilCancelled == 1) and true or nil
		dt.charges			= GetPlayerBuffApplications(buffIndex)
		dt.icon				= GetPlayerBuffTexture(buffIndex)
		
		table.insert(self.debuffdata, dt)
		i = i + 1
	end
	scan_happend.player = true
	updates_disabled = false
end

function ElkBuffBars:ScanData_TENCH()
	updates_disabled = true
	local maxtimes = self.db.account.maxtimes.TENCH
	for k, v in pairs(self.tenchdata) do
		RecycleDataTable(v)
		self.tenchdata[k] = nil
	end
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo()
	if ( hasMainHandEnchant ) then
		local timeleft = mainHandExpiration / 1000
		local timemax = timeleft
	
		local id = GetInventorySlotInfo("MainHandSlot")
		local name, rank = self:GetTempBuffName(id)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end
		local dt = GetDataTable()
		dt.id				= id
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "TENCH"
		dt.realtype			= "TENCH"
		dt.debufftype		= nil
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= nil
		dt.charges			= mainHandCharges
		dt.icon				= GetInventoryItemTexture("player", id)
		
		table.insert(self.tenchdata, dt)
	end
	if ( hasOffHandEnchant ) then
		local timeleft = offHandExpiration / 1000
		local timemax = timeleft

		local id = GetInventorySlotInfo("SecondaryHandSlot")
		local name, rank = self:GetTempBuffName(id)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end
		local dt = GetDataTable()
		dt.id				= id
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "TENCH"
		dt.realtype			= "TENCH"
		dt.debufftype		= nil
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= nil
		dt.charges			= mainHandCharges
		dt.icon				= GetInventoryItemTexture("player", id)
		
		table.insert(self.tenchdata, dt)
	end
	scan_happend.player = true
	updates_disabled = false
end

function ElkBuffBars:ScanData_TRACKING()
	updates_disabled = true
	for k, v in pairs(self.trackingdata) do
		RecycleDataTable(v)
		self.trackingdata[k] = nil
	end
	local icon = GetTrackingTexture()
	if icon then
		name = self:GetTrackingName()
		local dt = GetDataTable()
		dt.id				= id
		dt.name				= name
		dt.realname			= name
		dt.rank				= nil
		dt.type				= "TRACKING"
		dt.realtype			= "TRACKING"
		dt.debufftype		= nil
		dt.timeleft			= 1
		dt.timemax			= 1
		dt.untilcancelled	= true
		dt.charges			= 1
		dt.icon				= icon
		
		table.insert(self.trackingdata, dt)
	end
	scan_happend.player = true
	updates_disabled = false
end

function ElkBuffBars:ScanData_Unit_BUFF(unit)
	local unitvar = unit.."buffdata"
	if not self[unitvar] then return end
	updates_disabled = true
	for k, v in pairs(self[unitvar]) do
		RecycleDataTable(v)
		self[unitvar][k] = nil
	end
	local i = 1
	while true do
		local name, rank, icon, count = UnitBuff(unit, i)
		if not icon then break end
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end

		local dt = GetDataTable()
		dt.id				= buffIndex
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= unit:upper().."BUFF"
		dt.realtype			= unit:upper().."BUFF"
		dt.debufftype		= nil
		dt.timeleft			= 1
		dt.timemax			= 1
		dt.untilcancelled	= true
		dt.charges			= count
		dt.icon				= icon
		
		table.insert(self[unitvar], dt)
		i = i + 1
	end
	scan_happend[unit] = true
	updates_disabled = false
end

function ElkBuffBars:ScanData_Unit_DEBUFF(unit)
	local unitvar = unit.."debuffdata"
	if not self[unitvar] then return end
	updates_disabled = true
	for k, v in pairs(self[unitvar]) do
		RecycleDataTable(v)
		self[unitvar][k] = nil
	end
	local i = 1
	while true do
		local name, rank, icon, count, debuffType = UnitDebuff(unit, i);
		if not icon then break end
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end

		local dt = GetDataTable()
		dt.id				= buffIndex
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= unit:upper().."DEBUFF"
		dt.realtype			= unit:upper().."DEBUFF"
		dt.debufftype		= debuffType
		dt.timeleft			= 1
		dt.timemax			= 1
		dt.untilcancelled	= true
		dt.charges			= count
		dt.icon				= icon
		
		table.insert(self[unitvar], dt)
		i = i + 1
	end
	scan_happend[unit] = true
	updates_disabled = false
end

function ElkBuffBars:GetTempBuffName(id)
	gratuity:SetInventoryItem("player", id)
	local _, _, buffname = gratuity:Find("^(.+) %(%d+ [^%)]+%)$")
	if buffname then
		return buffname
	end
	local itemlink = GetInventoryItemLink("player", id)
	if itemlink then
		name = GetItemInfo(itemlink)
		return name or "Weapon "..id
	end
	return "Weapon "..id
end

function ElkBuffBars:GetTrackingName()
	gratuity:SetTrackingSpell()
	local name = gratuity:GetLine(1)
	return name or "Tracking..."
end

--
--
--
function ElkBuffBars:DoUpdates()
	if updates_disabled then return end
	
	for k, v in pairs(self.buffdata) do
		if not v.untilcancelled then
			local timeleft = GetPlayerBuffTimeLeft(v.id)
			if timeleft > v.timeleft then
				self:ScanData_BUFF()
				break
			end
			v.timeleft = timeleft
		end
	end

	for k, v in pairs(self.debuffdata) do
		if not v.untilcancelled then
			local timeleft = GetPlayerBuffTimeLeft(v.id)
			if timeleft > v.timeleft then
				self:ScanData_DEBUFF()
				break
			end
			v.timeleft = timeleft
		end
	end

	-- check for changed weaponbuffs
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo()
	if hasMainHandEnchant ~= old_hasMainHandEnchant or (mainHandExpiration or 0) > (old_mainHandExpiration or 0) or mainHandCharges ~= old_mainHandCharges
		or hasOffHandEnchant ~= old_hasOffHandEnchant or (offHandExpiration or 0) > (old_offHandExpiration or 0) or offHandCharges ~= old_offHandCharges then
		self:ScanData_TENCH()
	else
		for k, v in pairs(self.tenchdata) do
			if v.id == GetInventorySlotInfo("MainHandSlot") then
				v.timeleft = mainHandExpiration / 1000
			end
			if v.id == GetInventorySlotInfo("SecondaryHandSlot") then
				v.timeleft = offHandExpiration / 1000
			end
		end
	end
	old_hasMainHandEnchant = hasMainHandEnchant
	old_mainHandExpiration = mainHandExpiration
	old_mainHandCharges = mainHandCharges
	old_hasOffHandEnchant = hasOffHandEnchant
	old_offHandExpiration = offHandExpiration
	old_offHandCharges = offHandCharges
	
	-- if we did a scan, update bar data
	if scan_happend.player or scan_happend.pet or scan_happend.target then
		for k, v in pairs(self.bargroups) do
			v:UpdateData(scan_happend)
		end
		for k in pairs(scan_happend) do
			scan_happend[k] = false
		end
	else
		for k, v in pairs(self.bargroups) do
			v:UpdateTimeleft()
		end
	end
end


function ElkBuffBars:StickGroup(bargroup)
	local layout = bargroup.layout
	local id = layout.id
	local growup = layout.growup
	local container = bargroup:GetContainer()
	local base_y = growup and container:GetBottom() or container:GetTop()
	local base_left = container:GetLeft()
	local base_right = container:GetRight()
	layout.stickto = nil
	self:Print("Sticking bargroup", id)
	for k, v in pairs(self.bargroups) do
		local comp_container = v:GetContainer()
		local comp_y = growup and comp_container:GetTop() or comp_container:GetBottom()
		if math.abs(comp_y - base_y) < STICKTO_AREA then
			-- we are on the same y-area
			local rel_left = base_left - comp_container:GetLeft()
			local rel_right = base_right - comp_container:GetRight()
			if rel_left >= - STICKTO_AREA and rel_right <= STICKTO_AREA then
				self:Print(" - sticking to bargroup ", k)
				-- we are alo on the same x-area
				-- check if we would loop-stick to ourself
				local parent = v
				local hasloop = false
				while parent.layout.stickto do
					if parent.layout.stickto == id then
						self:Print(" - LOOP FOUND!")
						hasloop = true
						break
					end
					parent = self.bargroups[parent.layout.stickto]
				end
				if not hasloop then
					self:Print(" - SUCESS!")
					-- we have found a valid group to stick to
					layout.stickto = k
					if math.abs(rel_left) <= STICKTO_AREA and math.abs(rel_right) > STICKTO_AREA then
						layout.stickside = "LEFT"
					elseif math.abs(rel_left) > STICKTO_AREA and math.abs(rel_right) <= STICKTO_AREA then
						layout.stickside = "RIGHT"
					else
						layout.stickside = ""
					end
					container:ClearAllPoints()
					container:SetPoint((growup and "BOTTOM" or "TOP")..(layout.stickside or ""), comp_container, (growup and "TOP" or "BOTTOM")..(layout.stickside or ""))
					return true
				end
			end
		end
	end
	return false
end
